/*
**
**
**
** This file's purpose is to emulate the inner workings of a
** GameBoy Game Pak cartrige.
**
** All code is by Mark McGough. MadManMarkAu@hotmail.com
**
**
*/
#include <windows.h>
#include "commonIncludes.h"
#include "NRage PluginV2.h"
#include "PakIO.h"
#include "GBCart.h"

void ClearData(char *Data, int Length);

bool LoadCart(LPGBCART Cart, LPCTSTR RomFile, LPCTSTR RamFile, LPCTSTR TdfFile)
{
	HANDLE hRomFile;
	HANDLE hRamFile;
//	HANDLE hTdfFile;
	DWORD dwBytesRead;

	Cart->iCurrentRamBankNo = 0;
	Cart->iCurrentRomBankNo = 1;
	Cart->bRamEnableState = 1;

// Attempt to load the ROM file.
	hRomFile = CreateFile(RomFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL);
	if (hRomFile == INVALID_HANDLE_VALUE) {
		MessageBox( g_strEmuInfo.hMainWindow, "Can't open ROM file!", PLUGINERROR, MB_OK | MB_ICONERROR);
		return false;
	} else {
		ReadFile(hRomFile, &Cart->RomData[0], 0x40 * 0x4000, &dwBytesRead, NULL);
		CloseHandle(hRomFile);
	}

// Attempt to load the SRAM file.
	hRamFile = CreateFile(RamFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL);
	if (hRamFile == INVALID_HANDLE_VALUE) {
		MessageBox( g_strEmuInfo.hMainWindow, "Can't open SRAM file!\nContinuing with null savedata.", PLUGINWARNING, MB_OK | MB_ICONWARNING);
		ClearData(Cart->RamData, 0x08 * 0x2000);
	} else {
		ReadFile(hRamFile, &Cart->RamData[0], 0x08 * 0x2000, &dwBytesRead, NULL);
		CloseHandle(hRamFile);
	}

// TDF Files are currently unsoported, so we won't try to load one.
// When support is added, loading of the TDF goes in here.

	return true;
}

bool ReadCart(LPGBCART Cart, DWORD dwAddress, BYTE *Data)
{
	int i;

	DebugWriteOneline("** Cart. Read: ");
	DebugWriteWord(dwAddress);
	DebugWrite("");
	if ((dwAddress >= 0) && (dwAddress <= 0x3FFF)) {
		for (i=0; i<32; i++) Data[i] = Cart->RomData[dwAddress + i];
		DebugWrite("** Fixed ROM read");
	}
	if ((dwAddress >= 0x4000) && (dwAddress <= 0x7FFF)) {
		for (i=0; i<32; i++) Data[i] = Cart->RomData[(dwAddress - 0x4000) + i + (Cart->iCurrentRomBankNo * 0x4000)];
		DebugWriteOneline("** Banked ROM read:");
		DebugWriteByte(Cart->iCurrentRomBankNo);
		DebugWrite("");
	}
	if ((dwAddress >= 0xA000) && (dwAddress <= 0xBFFF)) {
		for (i=0; i<32; i++) Data[i] = Cart->RamData[(dwAddress - 0xA000) + i + (Cart->iCurrentRamBankNo * 0x2000)];
		DebugWriteOneline("** RAM read:");
		DebugWriteByte(Cart->iCurrentRamBankNo);
		DebugWrite("");
	}
	return true;
}

bool WriteCart(LPGBCART Cart, DWORD dwAddress, BYTE *Data)
{
	int i;

//	if ((dwAddress >= 0) && (dwAddress <= 0x1FFF)) { // RAM enable (Don't know what this does... Doesn't seem to affect a GB Rom cartrige)
//	}
	DebugWriteOneline("** Cart. Write: ");
	DebugWriteWord(dwAddress);
	DebugWrite("");

	if ((dwAddress >= 0x2000) && (dwAddress <= 0x3FFF)) { // ROM bank select
		Cart->iCurrentRomBankNo = Data[0] & 0x3F;
		DebugWriteOneline("** Set Rom:");
		DebugWriteByte(Cart->iCurrentRomBankNo);
		DebugWrite("");
	}
	if ((dwAddress >= 0x4000) && (dwAddress <= 0x5FFF)) { // RAM bank select
		Cart->iCurrentRamBankNo = Data[0] & 0x07;
		DebugWriteOneline("** Set Bank: ");
		DebugWriteByte(Cart->iCurrentRamBankNo);
		DebugWrite("");
	}
	if ((dwAddress >= 0xA000) && (dwAddress <= 0xBFFF)) { // Write to RAM
		DebugWriteOneline("** RAM write: ");
		DebugWriteByte(Cart->iCurrentRamBankNo);
		DebugWrite("");
		for (i=0; i<32; i++) Cart->RamData[(dwAddress - 0xA000) + i + (Cart->iCurrentRamBankNo * 0x2000)] = Data[i];
	}
	return true;
}

bool SaveCart(LPGBCART Cart, LPTSTR SaveFile, LPTSTR TimeFile)
{
	HANDLE hRamFile;
	DWORD dwBytesWritten;

	hRamFile = CreateFile(SaveFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL );
	WriteFile(hRamFile, Cart->RamData, 0x08 * 0x2000, &dwBytesWritten, NULL);
	CloseHandle(hRamFile);
	return true;
}

bool UnloadCart(LPGBCART Cart)
{
	// Nothing to do when unloading...
	return true;
}

// This is used to clear the RAM data to look like it has just been turned on.
// When a RAM chip is first turned on, it is filled with alternating 128-byte
// blocks of 0x00 and 0xFF.
void ClearData(char *Data, int Length)
{
	int i;

	for (i=0; i<Length; i++) {
		if ((i & 0x80) != 0x80) {
			Data[i] = (char) 0x00;
		} else {
			Data[i] = (char) 0xFF;
		}
	}
}
